*=======================================================================
      SUBROUTINE BOXCQP(B,X,D,XL,XR,AUX,N,ITRCNT,IQP,INP,INFO,COMACC,
     >		        IPR, SH, DD, Y, IP, E, WORK1, WORK2, LC,BP)
*=======================================================================
*  Desciption:                                                           *
*  BOXCQP: Convex quadratic solver s.t. simple bounds on the             *
*  variable.                                                             *
*                                                                        *
*  Date:                                                                 *
*        November, 2006                                                  *
*                                                                        *
*  Implemented by:                                                       *
*         C. Voglis                                                      *
*         Dept. of Computer Science, University of Ioannina, GREECE.     *
*         University of Ioannina, GREECE.                                *
*         Email: voglis@cs.uoi.gr                                        *
*                                                                        *
*         I. E. Lagaris                                                  *
*         Dept. of Computer Science, University of Ioannina, GREECE.     * 
*         Email: lagaris@cs.uoi.gr                                       * 
*                                                                        *
* Solves the box-constrained quadratic programming problem:              *	
*                                                                        *
*    minimize: 1/2*X^T*B*X + D*X  subject to:                            *
*    XL(I) <= X(I) <= XR(I) FOR  I=1,2,...,N                             *
*                                                                        *
*    where B is an nxn positive definite matrix.                         *
*    { dimensioned as B(N**2) }                                          *
*                                                                        *
*======================================================================  *
C  ---------------------------------------------------------------------
C  Input arguments:
C    N			Dimensionality of the problem.
C    B			NxN  matrix.
C    D			Nx1 right size vector
C    XL			Left margins.
C    XR			Right margins.
C    AUX		N*N auxiliary array.
C    IQP          Flag controlling the linear system solvers:
C			    IQP = 0  -> Use conjugate gradient to solve the 
C					produced linear systems. If the Newton point
C					is to be calculated (see INP) the first CGR
C					algorithm is prematurely stopped.
C			    IQP = 1  -> Use Cholesky decomposition to solve the 
C					produced linear systems.
C			    IQP = 2  -> Use a few conjugate gradient iterations for the 
C					Newton point approximation and then proceed with
C					Cholesky decomposition.
C   INP			Flag controlling the initial approximation.
C				INP = 1  ->	Begin with the Newton point xn = -B^{-1} d		
C				INP = 0  ->	Begin with x = xl, i.e. fix all variables 
C						    on lower bound. 
C   IPR			Printout level
C                 IPR = 0 -> No printout at all.
C                 IPR = 1 -> Simple printout: Prints ITRCNT, quadratic 
C						   function value. 
C                 IPR = 2 -> Full printout. Prints in addition the
C                            minimization parameters, contained in
C                            array X.
C
C    COMACC		Estimated machine accurancy.
C    SH			Nx1 work array.
C    DD			Nx1 work array.	
C    Y			Nx1 work array.
C    IP			Nx1 work array.
C    E			Nx1 work array.
C    WORK1		Nx1 work array.
C    WORK2		Nx1 work array.
C    LC			Nx1 work array.
C    BP			Nx1 work array.
C
C  Output:
C    X			On output the unique solution of the problem.
C    ITRCNT		Iterations performed. (Maximum value = N)
C    INFO			Reason for termination of this routine:
C				INFO = 1 -> Positive indefiniteness detected.
C				INFO = 2 -> Maximum number of iterations reached (N).			
C				INFO = 3 -> Algorithm converged to a solution.
C				INFO = 4 -> CGR: Detecting indefiniteness.			
C				INFO = 5 -> CGR: Large PBP number.
C
C  ---------------------------------------------------------------------
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DIMENSION B(N*N),X(N),D(N),AUX(N*N)
      DIMENSION XL(N),XR(N)
      DIMENSION SH(N), DD(N), Y(N), IP(N), E(N), WORK1(N), WORK2(N)
      DIMENSION LC(N),BP(N)
                 
      LOGICAL DEC
C
C  Calculate machine accuracy depended contants.
      TAU1 = COMACC ** (1./2.)
      TAU2 = COMACC ** (1./2.)
C
C Copy Hessian to an auxiliary work space.
      CALL COPYM(AUX, B, N)  
C
C Inverse sign of vector D(*) and prepare the initial
C CGR approximation.
       DO 11 I=1,N
         DD(I)=-D(I)
           X(I) = DD(I)/B((I-1)*N+I)
 11    CONTINUE
C
C Calculate Newton point. 
	IF (INP.EQ.1) THEN 
C
C Calculate the first iterate (newton point).     
       IF (IQP.EQ.0 .OR. IQP.EQ.2) THEN      
C
C Solve the system by using the conjugate gradient technique.
        CALL CGR (AUX,N,X,DD,N/10,K,COMACC,E,WORK1,WORK2, INFO)
	  IF (INFO.EQ.4.OR.INFO.EQ.5) THEN
	    RETURN
	  END IF 
       ELSE
C
C Solve the system by using the direct method (modified cholesky). 
	  CALL MODCHL(N,N,AUX,WORK1,COMACC,TAU1,TAU2,IP,E)
C
C Check for positive definiteness	
	 DO 753 II=1, N
	   IF (ABS(E(II)).GT.COMACC) THEN
            INFO = 1
	      RETURN 
	   END IF
 753   CONTINUE
C
C Call solve to obtain the solution of the initial quadratic problem.
	 CALL SOLVE(N,N,AUX,IP,X,DD,WORK1,WORK2)
       END IF
      ELSE
C
C Preffer a totally infeasible point. Let's say lower bound - 1.0.
	DO 12 I=1, N
	  X(I) = XL(I) - 1.0
12    CONTINUE
	END IF
C      
C Check which components fall out-of-bounds.
      LSC = 0
      DO 1 I=1,N
       IF (X(I).LT.XL(I))THEN
           LC(I) = -1
           X(I)=XL(I)
       ELSEIF (X(I).GT.XR(I))THEN
           LC(I) =  1
           X(I)=XR(I)
       ELSE
           LC(I) =  0
       ENDIF
C 
       IF (LC(I).EQ.0) LSC = LSC + 1
1     CONTINUE

       IF (LSC.EQ.N) THEN 
	  INFO = 2
        GOTO 100
       END IF
       
      ITRCNT = 0
C  ---------------------------------------------------------------------
C	Main loop begins here.
C  ---------------------------------------------------------------------
30    CONTINUE     
      ITRCNT = ITRCNT + 1 
      LSC = 0
      LLF = 0
      LRF = 0
      DO 31 I=1,N
       IF (LC(I).EQ.0) LSC = LSC + 1
       IF (LC(I).EQ.-1) LLF = LLF + 1
       IF (LC(I).EQ.1) LRF = LRF + 1
31    CONTINUE
C
C Print iteration stats 
      CALL PRINTO(B, D, X, N, ITRCNT, LLF, LRF, LSC,IPR) 
C	      
      IA = 0
      DO 2 I=1,N
       IF (LC(I).NE.0) GO TO 2
       IA = IA + 1
      JA = 0
      DO 3 J=1,N
       IF (LC(J).NE.0) GO TO 3
       JA = JA + 1
       INCO = (JA-1)*LSC+IA
       AUX(INCO)=B((J-1)*N+I)
3     CONTINUE
2     CONTINUE
      IA = 0
      DO 4 I=1,N
       IF (LC(I).NE.0) GO TO 4
        IA = IA + 1
        SH(IA)=D(I)
        Y(IA)=X(I)
      DO 5 J=1,N
        SH(IA) = SH(IA)+SPLIT(LC(J),XL(J),XR(J))*B((J-1)*N+I)
5     CONTINUE
        SH(IA) = -SH(IA)
4     CONTINUE

C
C If LSC.EQ.0 there is no linear system to solve.
	IF (LSC.LE.0) GOTO 8
      IF (IQP.EQ.0) THEN   
       CALL CGR (AUX,LSC,Y,SH,2*LSC,K,COMACC,E,WORK1,WORK2,INFO)
	 IF (INFO.EQ.4.OR.INFO.EQ.5) THEN
	    RETURN
	  END IF 
      ELSE
C
C Solve the system by using the direct method (modified cholesky).    
	CALL MODCHL(LSC,LSC,AUX,WORK1,COMACC,TAU1,TAU2,IP,E)
C
C Check mildly for positive definites 
	IF (ABS(E(LSC)).GE.COMACC) THEN 	 
	 INFO = 1 
	 RETURN 
	END IF
C
C Solve the reduce system on LSC variables. 
	CALL SOLVE(LSC,LSC,AUX,IP,Y,SH,WORK1,WORK2)
      END IF 
8     CONTINUE
      IA = 0
      DO 6 I=1,N
      IF (LC(I).EQ.0) THEN
        IA = IA + 1
        X(I)=Y(IA)
      ENDIF
6     CONTINUE
      CALL MATVEC(B,N,X,BP)
      DO 10 I = 1,N
       BP(I) = BP(I)+D(I)
10    CONTINUE
      DEC = .FALSE.
      DO 7 I=1,N

      IF (LC(I).EQ.-1 .AND.  BP(I) .LE.0)THEN
          LC(I)=0
          DEC = .TRUE.
      ELSEIF (LC(I).EQ. 1 .AND. -BP(I) .LE.0) THEN
          LC(I)=0
          DEC = .TRUE.
      ELSE
          IF(X(I).LT.XL(I)) THEN 
              X(I)=XL(I)
              LC(I)=-1
          DEC = .TRUE.
          ELSEIF (X(I).GT.XR(I)) THEN
              X(I)=XR(I)
              LC(I)= 1
          DEC = .TRUE.
          ENDIF
      ENDIF
7     CONTINUE
C
999   CONTINUE
C
C Termination: Too  many iterations.
      IF (ITRCNT.GE.(N/4)) GO TO 100
      IF (DEC) GO TO 30
C  ---------------------------------------------------------------------
C	Main loop ends here.
C  ---------------------------------------------------------------------
100   CONTINUE
	INFO = 3

       END

C  ---------------------------------------------------------------------
      SUBROUTINE SCAPRO(X,Y,N,SP)
C  ---------------------------------------------------------------------
C
C  Description:
C    Calculates the scalar product SP = x'y
C
C  ---------------------------------------------------------------------
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DIMENSION X(N),Y(N)
      SP = 0.
      DO 1 I=1,N
        SP = SP + X(I)*Y(I)
1     CONTINUE
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MATVEC(B,N,P,BP)
C  ---------------------------------------------------------------------
C
C  Description:
C    Calculates the product BP = B*p, B being an  NxN matrix
C    P is a vector in N-dimensions.
C
C  ---------------------------------------------------------------------
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DIMENSION B(N,N),P(N),BP(N)
      DO 1 I = 1,N
        BP(I)=0.
        DO 2 J = 1,N
         BP(I) = BP(I) + B(I,J)*P(J)
2       CONTINUE
1     CONTINUE
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CGR (B,N,X,D,MM,K,COMACC,P,R,BP,INFO)
C  ---------------------------------------------------------------------
C
C  Description:
C    Solves the linear system [B]*x=d, where [B] is an nxn symmetric matrix
C    using the method of conjugate gradient.
C
C  Input arguments:
C    N			Dimensionality of the problem.
C    B			NxN matrix.
C    D			Nx1 right size vector
C    MM			Maximum nmber of CGR iterations.
C    COMACC		Estimated machine accurancy.
C    P			Nx1 work array.
C    R			Nx1 work array.
C    BP			Nx1 work array.

C
C  Output:
C    K			CGR iterations performed. K = -1, means 
C				that MM iterations were performed without 
C				achieving convergence.
C    INFO			Reason for termination of this routine:
C				INFO = 3 -> Normal termination. 
C				INFO = 4 -> B singular.			
C				INFO = 5 -> Large denominator p'Bp.
C
C  Input/Output arguments:
C    X			On input the initial approximation of the solution.
C				On output the CGR approximantion of the solution
C				(after MM iterations).
C  ---------------------------------------------------------------------
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DIMENSION B(N,N),X(N),D(N)
      DIMENSION P(N),R(N),BP(N)
      BIGNUM = 1.D+300
	ETA = SQRT(COMACC)
      CALL MATVEC(B,N,X,R)
      DO 1 I=1,N
       R(I) = R(I)-D(I)
       P(I) = -R(I)
1     CONTINUE
      K = 0
30    CONTINUE
      CALL SCAPRO(R,R,N,RR) 
C
C  Stopping condition.
      IF (RR .LT. ETA) THEN
	 INFO = 3
	 RETURN
	END IF 
      CALL MATVEC(B,N,P,BP)
      CALL SCAPRO(BP,P,N,PBP)      
      IF (PBP.EQ.0) THEN
	  INFO = 4 
	  RETURN 
      ENDIF

      IF (PBP.GT.BIGNUM) THEN
	  INFO = 5
	  RETURN 
      END IF 

       AL = RR/PBP
       DO 2 I =1,N
         X(I) = X(I) + AL*P(I)
         R(I) = R(I) + AL*BP(I)
2      CONTINUE
       CALL SCAPRO(R,R,N,RR1)
       BETA = RR1/RR
       DO 3 I = 1,N
        P(I) = -R(I)+BETA*P(I)
3      CONTINUE
       K = K+1
       IF (K.GT.MM) THEN
          K=-1
          RETURN
       ENDIF
       GO TO 30
      END
C  ---------------------------------------------------------------------
      FUNCTION SPLIT(L,A,B)
C  ---------------------------------------------------------------------
C
C  Description:
C    Function SPLIT returns A, B, or 0 according to the value of L.
C
C  Input arguments:
C    L			Matrix to be copied.
C    A			Matrix dimension.
C    B
C
C  Return value:
C    A			If L = -1.
C    B			If L = +1.
C    0			If L = 0
C  ---------------------------------------------------------------------

      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
	IF (L.EQ.1) THEN 
	  SPLIT = B
	ELSE IF (L.EQ.-1) THEN
	  SPLIT = A
	ELSE 
	  SPLIT = 0.0
	END IF
      END


C  ---------------------------------------------------------------------      
      SUBROUTINE COPYM(A, B, N)
C  ---------------------------------------------------------------------
C
C  Description:
C    Copies all the elements of matrix B to matrix A.
C
C  Input arguments:
C    B			Matrix to be copied.
C    N			Matrix dimension.
C
C  Output arguments:
C    A			Matrix to be written.
C
C  ---------------------------------------------------------------------
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)          
      DIMENSION A(N*N), B(N*N)
      DO 10  I=1, N*N
       A(I) = B(I)
10    CONTINUE      
      END 

C  ---------------------------------------------------------------------
      SUBROUTINE BOXCQPMS ( INFO, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Constructs a message, describing the reason for termination of the
C    BOXCQP method.
C
C  Input arguments:
C    INFO       Reason for termination of the BOXCQP method, as returned
C               by subroutine BOXCQP.
C
C  Output arguments:
C    MESS       Message, describing the reason for termination.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C
      IF (INFO.EQ.1) THEN
         MESS = 'Positive indefiniteness detected '
      ELSE IF (INFO.EQ.2) THEN
         MESS = 'Maximum number of iterations reached (N)'
      ELSE IF (INFO.EQ.3) THEN
         MESS = 'Algorithm converged to a solution.'
      ELSE IF (INFO.EQ.4) THEN
         MESS = 'CGR: Detecting indefiniteness.'
      ELSE IF (INFO.EQ.5) THEN
         MESS = 'CGR: Large PBP number.'
      ELSE
         MESS = 'Wrong INFO number. '
      END IF
      END
C  ---------------------------------------------------------------------
	SUBROUTINE MACACC (COMACC)
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine determines the relative machine accuracy (COMACC). 
C  ---------------------------------------------------------------------
	IMPLICIT DOUBLE PRECISION (A-H, O-Z)
	TEN = 10.
	NDRTF = 0
      IF (NDRTF.EQ.0) THEN
         ONE = 1.
         DO 10,I=1,1000
            COMACC = TEN**(-I)	
            X1 = ONE+COMACC
            X2 = ONE-COMACC
            IF (X1.EQ.ONE .OR. X2.EQ.ONE) THEN
               NDRTF = I-1
               GOTO 20
            END IF
10       CONTINUE
         WRITE (*, *) 'Cannot determine accuracy.'
	   COMACC = -1.0
	   RETURN 
      END IF
20	CONTINUE 
	END
C  ---------------------------------------------------------------------
	SUBROUTINE PRINTO(B, D, X, N, ITRCNT, LLF, LRF, LSC,IPR)      
C  ---------------------------------------------------------------------
C  Description:
C    Issues informative messages during a minimization session.
C    Normally, it is called once per iteration.
C
C  Input arguments:
C    B	      NxN Hessian matrix of the quadratic
C    D		  Nx1 Linear term vector.
C    X		  Nx1 current feasible iteration.
C    ITRCNT     Number of iterations that have been performed by the
C               calling routine.
C    LLF		  Number of variables on the lower bound.
C    LRF		  Number of variables on the upper bound.
C    LSC        Number of variables strictly feasible.
C    IPR        Printout level:
C               IPR = 0 -> Nothing is printed.
C               IPR = 1 -> Simple printout: Prints ITRCNT, quadratic 
C						 function value. 
C               IPR = 2 -> Full printout. Prints in addition the
C                          minimization parameters, contained in
C                          array X.
C
C  ---------------------------------------------------------------------
	IMPLICIT DOUBLE PRECISION (A-H, O-Z)
	DIMENSION B(N, N), D(N), X(N) 
	CHARACTER*10 C1
C
	IF (IPR.EQ.1) THEN 
	 CALL I2STR(ITRCNT,C1,LE1)
	 WRITE (*, *) '--------------------------------------------'
	 WRITE (*, 14) C1, QUADRATIC(B, D, X, N)
	 WRITE (*, *) '--------------------------------------------'
	END IF
	IF (IPR.EQ.2) THEN
	 CALL I2STR(ITRCNT,C1,LE1) 
	 WRITE (*, *) '--------------------------------------------'
	 WRITE (*, 14) C1, QUADRATIC(B, D, X, N)
       WRITE(*,*) '# of vars on lower bound : ',LLF
       WRITE(*,*) '# of vars on upper bound : ',LRF
	 WRITE(*,*) '# of vars inside the box : ',LSC
	 DO 10 I=1, N
		WRITE (*, *) I, ' = ', X(I)
10     CONTINUE 
	 WRITE (*, *) '--------------------------------------------'
	END IF
14    FORMAT(' Iter: ', A,  'Function val: ', F14.5)
	END 
C  ---------------------------------------------------------------------
	FUNCTION QUADRATIC(B, D, X, N)
C  ---------------------------------------------------------------------
C  Description:
C    Evaluates the quadratic function 1/2 x' B x + x' d at x.
C  ---------------------------------------------------------------------
	IMPLICIT DOUBLE PRECISION (A-H, O-Z)
	DIMENSION B(N, N), D(N), X(N)
	 RES = 0.0 
	 DO 10 I=1, N 
	  DO 20 J=1, N
		RES = RES + B(I, J)*X(I)*X(J)
20      CONTINUE 
10     CONTINUE
	 RES = 0.5 * RES 

	 DO 30 I=1, N
	   RES = RES + D(I)*X(I) 
30     CONTINUE 
	
	QUADRATIC = RES
	END 
C  ---------------------------------------------------------------------
      SUBROUTINE I2STR ( N, STR, LES )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts an integer (N) to its string representation (STR).
C
C  Input arguments:
C    N          The integer to be converted.
C
C  Output arguments:
C    STR        The first LES characters of STR contain the string
C               representation of N.
C    LES        Effective length of STR.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*5 FORM
      DATA FORM / '(I__)' /
C  ---------------------------------------------------------------------
C
      LE = LEN(STR)
      WRITE (FORM(3:4),'(I2.2)') LE
      WRITE (STR,FORM) N
C
      DO 10,I=1,LE
         IF (STR(I:I).NE.' ') THEN
            DO 20,J=I,LE
               JI = J-I+1
               STR(JI:JI) = STR(J:J)
20          CONTINUE
            STR(LE-I+2:) = ' '
            LES = LE-I+1
            RETURN
         END IF
10    CONTINUE
      LES = LE
C
      END

