C    --- MODIFIED CHOLESKY
C
C*********************************************************************
C
C       SUBROUTINE NAME: MODCHL
C
C       AUTHORS :  ELIZABETH ESKOW AND ROBERT B. SCHNABEL
C
C       DATE    : DECEMBER, 1988
C
C       PURPOSE : PERFORM A MODIFIED CHOLESKY FACTORIZATION
C                 OF THE FORM (PTRANSPOSE)AP  + E = L(LTRANSPOSE),
C       WHERE L IS STORED IN THE LOWER TRIANGLE OF THE
C       ORIGINAL MATRIX A.
C       THE FACTORIZATION HAS 2 PHASES:
C        PHASE 1: PIVOT ON THE MAXIMUM DIAGONAL ELEMENT.
C            CHECK THAT THE NORMAL CHOLESKY UPDATE
C            WOULD RESULT IN A POSITIVE DIAGONAL
C            AT THE CURRENT ITERATION, AND
C            IF SO, DO THE NORMAL CHOLESKY UPDATE,
C            OTHERWISE SWITCH TO PHASE 2.
C        PHASE 2: PIVOT ON THE MINIMUM OF THE NEGATIVES
C            OF THE LOWER GERSCHGORIN BOUND
C            ESTIMATES.
C            COMPUTE THE AMOUNT TO ADD TO THE
C            PIVOT ELEMENT AND ADD THIS
C            TO THE PIVOT ELEMENT.
C            DO THE CHOLESKY UPDATE.
C            UPDATE THE ESTIMATES OF THE
C            GERSCHGORIN BOUNDS.
C
C       INPUT   : NDIM    - LARGEST DIMENSION OF MATRIX THAT
C                           WILL BE USED
C
C                 N       - DIMENSION OF MATRIX A
C
C                 A       - N*N SYMMETRIC MATRIX (ONLY LOWER TRIANGULAR
C            PORTION OF A, INCLUDING THE MAIN DIAGONAL, IS USED)
C
C                 G       - N*1 WORK ARRAY
C
C                 MCHEPS - MACHINE PRECISION
C
C                TAU1    - TOLERANCE USED FOR DETERMINING WHEN TO SWITCH
C                          PHASE 2
C
C                TAU2    - TOLERANCE USED FOR DETERMINING THE MAXIMUM
C                          CONDITION NUMBER OF THE FINAL 2X2 SUBMATRIX.
C
C
C       OUTPUT  : L     - STORED IN THE MATRIX A (IN LOWER TRIANGULAR
C                           PORTION OF A, INCLUDING THE MAIN DIAGONAL)
C
C                 P     - A RECORD OF HOW THE ROWS AND COLUMNS
C                         OF THE MATRIX WERE PERMUTED WHILE
C                         PERFORMING THE DECOMPOSITION
C
C                 E     - N*1 ARRAY, THE ITH ELEMENT IS THE
C                         AMOUNT ADDED TO THE DIAGONAL OF A
C                         AT THE ITH ITERATION
C
C
C***********************************************************************
      SUBROUTINE MODCHL(NDIM,N,A,G,MCHEPS,TAU1,TAU2,P,E)
*
      INTEGER N,NDIM
      DOUBLE PRECISION A(NDIM,N),G(N),MCHEPS,TAU1,TAU2
      INTEGER P(N)
      DOUBLE PRECISION E(N)
*
C
C  J        - CURRENT ITERATION NUMBER
C  IMING    - INDEX OF THE ROW WITH THE MIN. OF THE
C           NEG. LOWER GERSCH. BOUNDS
C  IMAXD    - INDEX OF THE ROW WITH THE MAXIMUM DIAG.
C           ELEMENT
C  I,ITEMP,JPL,K  - TEMPORARY INTEGER VARIABLES
C  DELTA    - AMOUNT TO ADD TO AJJ AT THE JTH ITERATION
C  GAMMA    - THE MAXIMUM DIAGONAL ELEMENT OF THE ORIGINAL
C           MATRIX A.
C  NORMJ    - THE 1 NORM OF A(COLJ), ROWS J+1 --> N.
C  MING     - THE MINIMUM OF THE NEG. LOWER GERSCH. BOUNDS
C  MAXD     - THE MAXIMUM DIAGONAL ELEMENT
C  TAUGAM - TAU1 * GAMMA
C  PHASE1      - LOGICAL, TRUE IF IN PHASE1, OTHERWISE FALSE
C  DELTA1,TEMP,JDMIN,TDMIN,TEMPJJ - TEMPORARY DOUBLE PRECISION VARS.
C
*
      INTEGER J,IMING,I,IMAXD,ITEMP,JP1,K
      DOUBLE PRECISION DELTA,GAMMA
      DOUBLE PRECISION NORMJ, MING,MAXD
      DOUBLE PRECISION DELTA1,TEMP,JDMIN,TDMIN,TAUGAM,TEMPJJ
      LOGICAL PHASE1
      INTRINSIC ABS, MAX, SQRT, MIN
      
*
      CALL MYINIT(N, NDIM, A, PHASE1, DELTA, P, G, E,
     *         MING,TAU1,GAMMA,TAUGAM)
C
C     CHECK FOR N=1
C
      IF (N.EQ.1) THEN
         DELTA = (TAU2 * ABS(A(1,1))) - A(1,1)
         IF (DELTA .GT. 0) E(1) = DELTA
         IF (A(1,1) .EQ. 0) E(1) = TAU2
         A(1,1)=SQRT(A(1,1)+E(1))
      ENDIF
C
      DO 200 J = 1, N-1
C
C        PHASE 1
C
         IF ( PHASE1 ) THEN
C
C           FIND INDEX OF MAXIMUM DIAGONAL ELEMENT A(I,I) WHERE I>=J
C
            MAXD = A(J,J)
            IMAXD = J
            DO 20 I = J+1, N
               IF (MAXD .LT. A(I,I)) THEN
                  MAXD = A(I,I)
                  IMAXD = I
               END IF
 20         CONTINUE
*
C
C           PIVOT TO THE TOP THE ROW AND COLUMN WITH THE MAX DIAG
C
            IF (IMAXD .NE. J) THEN
C
C              SWAP ROW J WITH ROW OF MAX DIAG
C
               DO 30 I = 1, J-1
                  TEMP = A(J,I)
                  A(J,I) = A(IMAXD,I)
                  A(IMAXD,I) = TEMP
 30            CONTINUE
C
C              SWAP COLJ AND ROW MAXDIAG BETWEEN J AND MAXDIAG
C
               DO 35 I = J+1,IMAXD-1
                  TEMP = A(I,J)
                  A(I,J) = A(IMAXD,I)
                  A(IMAXD,I) = TEMP
 35            CONTINUE
C
C              SWAP COLUMN J WITH COLUMN OF MAX DIAG
C
               DO 40 I = IMAXD+1, N
                  TEMP = A(I,J)
                  A(I,J) = A(I,IMAXD)
                  A(I,IMAXD) = TEMP
 40            CONTINUE
C
C              SWAP DIAG ELEMENTS
C
               TEMP = A(J,J)
               A(J,J) = A(IMAXD,IMAXD)
               A(IMAXD,IMAXD) = TEMP
C
C              SWAP ELEMENTS OF THE PERMUTATION VECTOR
C
               ITEMP = P(J)
               P(J) = P(IMAXD)
               P(IMAXD) = ITEMP
*
            END IF
*
*
C           CHECK TO SEE WHETHER THE NORMAL CHOLESKY UPDATE FOR THIS
C           ITERATION WOULD RESULT IN A POSITIVE DIAGONAL,
C           AND IF NOT THEN SWITCH TO PHASE 2.
*
            JP1 = J+1
            TEMPJJ=A(J,J)
*
C ALLAXA ALLAXA
C
            IF (TEMPJJ.GT.0) THEN
*
               JDMIN=A(JP1,JP1)
               DO 60 I = JP1, N
                  TEMP = A(I,J) * A(I,J) / TEMPJJ
                  TDMIN = A(I,I) - TEMP
                  JDMIN = MIN(JDMIN, TDMIN)
 60            CONTINUE
*
               IF (JDMIN .LT. TAUGAM) THEN 
	         PHASE1 = .FALSE.
C		 PRINT *, 'PHASE 2 TAUGAM'
   	       END IF	 
*
            ELSE
*
               PHASE1 = .FALSE.
C               PRINT *, 'PHASE 2 NEGATIVE'
*
            END IF
*
            IF (PHASE1) THEN
C
C              DO THE NORMAL CHOLESKY UPDATE IF STILL IN PHASE 1
C
               A(J,J) = SQRT(A(J,J))
               TEMPJJ = A(J,J)
               DO 70 I = JP1, N
                  A(I,J) = A(I,J) / TEMPJJ
 70            CONTINUE
               DO 80 I=JP1,N
                  TEMP=A(I,J)
                  DO 75 K = JP1, I
                     A(I,K) = A(I,K) - (TEMP * A(K,J))
 75               CONTINUE
 80            CONTINUE
*
               IF (J .EQ. N-1) A(N,N)=SQRT(A(N,N))
*
            ELSE
*
C
C              CALCULATE THE NEGATIVES OF THE LOWER GERSCHGORIN BOUNDS
C
               CALL GERSCH(NDIM,N,A,J,G)
*
            END IF
*
         END IF
*
*
C
C        PHASE 2
C
         IF (.NOT. PHASE1) THEN
*
            IF (J .NE. N-1) THEN
C
C              FIND THE MINIMUM NEGATIVE GERSHGORIN BOUND
C
*
               IMING=J
               MING = G(J)
               DO 90 I = J+1,N
                  IF (MING .GT. G(I)) THEN
                     MING = G(I)
                     IMING = I
                  END IF
 90            CONTINUE
*
C
C               PIVOT TO THE TOP THE ROW AND COLUMN WITH THE
C               MINIMUM NEGATIVE GERSCHGORIN BOUND
C
                IF (IMING .NE. J) THEN
C
C                  SWAP ROW J WITH ROW OF MIN GERSCH BOUND
C
                   DO 100 I = 1, J-1
                      TEMP = A(J,I)
                       A(J,I) = A(IMING,I)
                       A(IMING,I) = TEMP
 100               CONTINUE
C
C                  SWAP COLJ WITH ROW IMING FROM J TO IMING
C
                   DO 105 I = J+1,IMING-1
                      TEMP = A(I,J)
                      A(I,J) = A(IMING,I)
                      A(IMING,I) = TEMP
 105              CONTINUE
C
C                 SWAP COLUMN J WITH COLUMN OF MIN GERSCH BOUND
C
                  DO 110 I = IMING+1, N
                     TEMP = A(I,J)
                     A(I,J) = A(I,IMING)
                     A(I,IMING) = TEMP
 110              CONTINUE
C
C                 SWAP DIAGONAL ELEMENTS
C
                  TEMP = A(J,J)
                  A(J,J) = A(IMING,IMING)
                  A(IMING,IMING) = TEMP
C
C                 SWAP ELEMENTS OF THE PERMUTATION VECTOR
C
                  ITEMP = P(J)
                  P(J) = P(IMING)
                  P(IMING) = ITEMP
C
C                 SWAP ELEMENTS OF THE NEGATIVE GERSCHGORIN BOUNDS VECTO
C
                  TEMP = G(J)
                  G(J) = G(IMING)
                  G(IMING) = TEMP
*
               END IF
C
C              CALCULATE DELTA AND ADD TO THE DIAGONAL.
C              DELTA=MAX{0,-A(J,J) + MAX{NORMJ,TAUGAM},DELTA_PREVIOUS}
C              WHERE NORMJ=SUM OF |A(I,J)|,FOR I=1,N,
C              DELTA_PREVIOUS IS THE DELTA COMPUTED AT THE PREVIOUS ITER
C              AND TAUGAM IS TAU1*GAMMA.
C
*
               NORMJ = 0.0
               DO 140 I = J+1, N
                  NORMJ = NORMJ + ABS(A(I,J))
 140           CONTINUE
*
               TEMP = MAX(NORMJ,TAUGAM)
               DELTA1 = TEMP - A(J,J)
               TEMP = 0.0
               DELTA1 = MAX(TEMP, DELTA1)
               DELTA = MAX(DELTA1,DELTA)
               E(J) =  DELTA
               A(J,J) = A(J,J) + E(J)

C
C              UPDATE THE GERSCHGORIN BOUND ESTIMATES
C              (NOTE: G(I) IS THE NEGATIVE OF THE
C               GERSCHGORIN LOWER BOUND.)
C
               IF (A(J,J) .NE. NORMJ) THEN
                  TEMP = (NORMJ/A(J,J)) - 1.0
*
                  DO 150 I = J+1, N
                     G(I) = G(I) + ABS(A(I,J)) * TEMP
 150              CONTINUE
*
               END IF
C
C              DO THE CHOLESKY UPDATE
C
               A(J,J) = SQRT(A(J,J))
               TEMPJJ = A(J,J)
               DO 160 I = J+1, N
                  A(I,J) = A(I,J) / TEMPJJ
 160           CONTINUE
               DO 180 I = J+1, N
                  TEMP = A(I,J)
                  DO 170 K = J+1, I
                     A(I,K) = A(I,K) - (TEMP * A(K,J))
 170              CONTINUE
 180           CONTINUE
*
            ELSE
*
               CALL FIN2X2(NDIM, N, A, E, J, TAU2, DELTA,GAMMA)
*
            END IF
*
         END IF
*
 200   CONTINUE
*
      RETURN
      END
C***********************************************************************
C       SUBROUTINE NAME : MYINIT
C
C       PURPOSE : SET UP FOR START OF CHOLESKY FACTORIZATION
C
C       INPUT : N, NDIM, A, TAU1
C
C       OUTPUT : PHASE1    - BOOLEAN VALUE SET TO TRUE IF IN PHASE ONE,
C             OTHERWISE FALSE.
C      DELTA     - AMOUNT TO ADD TO AJJ AT ITERATION J
C      P,G,E - DESCRIBED ABOVE IN MODCHL
C      MING      - THE MINIMUM NEGATIVE GERSCHGORIN BOUND
C      GAMMA     - THE MAXIMUM DIAGONAL ELEMENT OF A
C      TAUGAM  - TAU1 * GAMMA
C
C***********************************************************************
      SUBROUTINE MYINIT(N,NDIM,A,PHASE1,DELTA,P,G,E,MING,
     *                TAU1,GAMMA,TAUGAM)
*
      INTEGER N,NDIM
      DOUBLE PRECISION A(NDIM,N)
      LOGICAL PHASE1
      DOUBLE PRECISION DELTA,G(N),E(N)
      INTEGER P(N)
      DOUBLE PRECISION MING,TAU1,GAMMA,TAUGAM
      INTRINSIC ABS, MAX
*
*
      PHASE1 = .TRUE.
      DELTA = 0.0
      MING = 0.0
      DO 10 I=1,N
         P(I)=I
         G(I)= 0.0
         E(I) = 0.0
 10   CONTINUE
*
C
C     FIND THE MAXIMUM MAGNITUDE OF THE DIAGONAL ELEMENTS.
C     IF ANY DIAGONAL ELEMENT IS NEGATIVE, THEN PHASE1 IS FALSE.
C
      GAMMA = 0.0
      DO 20 I=1,N
         GAMMA=MAX(GAMMA,ABS(A(I,I)))
         IF (A(I,I) .LT. 0.0) PHASE1 = .FALSE.
 20   CONTINUE
*
      TAUGAM = TAU1 * GAMMA
*
C
C     IF NOT IN PHASE1, THEN CALCULATE THE INITIAL GERSCHGORIN BOUNDS
C     NEEDED FOR THE START OF PHASE2.
C
      IF ( .NOT.PHASE1) CALL GERSCH(NDIM,N,A,1,G)
*
      RETURN
      END
C***********************************************************************
C
C       SUBROUTINE NAME : GERSCH
C
C       PURPOSE : CALCULATE THE NEGATIVE OF THE GERSCHGORIN BOUNDS
C                 CALLED ONCE AT THE START OF PHASE II.
C
C       INPUT   : NDIM, N, A, J
C
C       OUTPUT  : G - AN N VECTOR CONTAINING THE NEGATIVES OF THE
C           GERSCHGORIN BOUNDS.
C
C***********************************************************************
      SUBROUTINE GERSCH(NDIM, N, A, J, G)
*
      INTEGER NDIM, N, J
      DOUBLE PRECISION A(NDIM,N), G(N)
*
      INTEGER I, K
      DOUBLE PRECISION OFFROW
      INTRINSIC ABS
*
      DO 30 I = J, N
         OFFROW = 0.0
         DO 10 K = J, I-1
            OFFROW = OFFROW + ABS(A(I,K))
 10      CONTINUE
         DO 20 K = I+1, N
            OFFROW = OFFROW + ABS(A(K,I))
 20      CONTINUE
            G(I) = OFFROW - A(I,I)
 30   CONTINUE
*
      RETURN
      END
C***********************************************************************
C
C  SUBROUTINE NAME : FIN2X2
C
C  PURPOSE : HANDLES FINAL 2X2 SUBMATRIX IN PHASE II.
C            FINDS EIGENVALUES OF FINAL 2 BY 2 SUBMATRIX,
C            CALCULATES THE AMOUNT TO ADD TO THE DIAGONAL,
C            ADDS TO THE FINAL 2 DIAGONAL ELEMENTS,
C            AND DOES THE FINAL UPDATE.
C
C  INPUT : NDIM, N, A, E, J, TAU2,
C          DELTA - AMOUNT ADDED TO THE DIAGONAL IN THE
C                  PREVIOUS ITERATION
C
C  OUTPUT : A - MATRIX WITH COMPLETE L FACTOR IN THE LOWER TRIANGLE,
C           E - N*1 VECTOR CONTAINING THE AMOUNT ADDED TO THE DIAGONAL
C               AT EACH ITERATION,
C           DELTA - AMOUNT ADDED TO DIAGONAL ELEMENTS N-1 AND N.
C
C***********************************************************************
      SUBROUTINE FIN2X2(NDIM, N, A, E, J, TAU2, DELTA,GAMMA)
*
      INTEGER NDIM, N, J
      DOUBLE PRECISION A(NDIM,N), E(N), TAU2, DELTA,GAMMA
*
      DOUBLE PRECISION T1, T2, T3,LMBD1,LMBD2,LMBDHI,LMBDLO
      DOUBLE PRECISION DELTA1, TEMP
      INTRINSIC SQRT, MAX, MIN
*
C
C     FIND EIGENVALUES OF FINAL 2 BY 2 SUBMATRIX
C
      T1 = A(N-1,N-1) + A(N,N)
      T2 = A(N-1,N-1) - A(N,N)
      T3 = SQRT(T2*T2 + 4.0*A(N,N-1)*A(N,N-1))
      LMBD1 = (T1 - T3)/2.
      LMBD2 = (T1 + T3)/2.
      LMBDHI = MAX(LMBD1,LMBD2)
      LMBDLO = MIN(LMBD1,LMBD2)
C
C     FIND DELTA SUCH THAT:
C     1.  THE L2 CONDITION NUMBER OF THE FINAL
C     2X2 SUBMATRIX + DELTA*I <= TAU2
C     2. DELTA >= PREVIOUS DELTA,
C     3. LMBDLO + DELTA >= TAU2 * GAMMA,
C     WHERE LMBDLO IS THE SMALLEST EIGENVALUE OF THE FINAL
C     2X2 SUBMATRIX
C
*
      DELTA1=(LMBDHI-LMBDLO)/(1.0-TAU2)
      DELTA1= MAX(DELTA1,GAMMA)
      DELTA1= TAU2 * DELTA1 - LMBDLO
      TEMP = 0.0
      DELTA = MAX(DELTA, TEMP)
      DELTA = MAX(DELTA, DELTA1)
*
      IF (DELTA .GT. 0.0) THEN
         A(N-1,N-1) = A(N-1,N-1) + DELTA
         A(N,N) = A(N,N) + DELTA
         E(N-1) = DELTA
         E(N) = DELTA
      END IF
C
C     FINAL UPDATE
C
      A(N-1,N-1) = SQRT(A(N-1,N-1))
      A(N,N-1) = A(N,N-1)/A(N-1,N-1)
      A(N,N) = A(N,N) - (A(N,N-1)**2)
      A(N,N) = SQRT(A(N,N))
*
      RETURN
      END
C**********************************************************************
C
C     SUBROUTINE NAME : SOLVE
C
C     PURPOSE :  SOLVES (LLTRANSPOSE)P(X)=P(B), WHERE L IS STORED IN
C                THE LOWER TRIANGLE OF A, P IS THE RECORD OF THE
C                PERMUTATIONS  PERFORMED IN FORMING L,B IS THE RHS VECTO
C                AND X IS THE RESULT.
C                L IS THE RESULT OF THE MODIFIED CHOLESKY FACTORIZATION
C                WHICH COMPUTES PTRANSPOSE(A+E)P=LLTRANSPOSE.
C
C  INPUT:   NDIM  - LARGEST DIMENSION OF MATRIX THAT WILL BE USED
C
C           N     - DIMENSION OF MATRIX
C
C           A     - N*N ARRAY (ONLY LOWER TRIANGLE PORTION,
C                   INCLUDING THE DIAGONAL, IS USED).
C
C           P     - N*1 INTEGER VECTOR THAT CONTAINS A RECORD
C                   OF PERMUTATIONS PERFORMED IN FORMING L.
C                   I.E. EACH PI WAS INITIALIZED TO I, AND IF
C                   ROW AND COLUMNS I & J WERE INTERCHANGED WHEN
C                   COMPUTING L, THE VALUES OF PI AND PJ WERE SWAPPED.
C
C           B     - N*1 DOUBLE PRECISION VECTOR WHICH IS THE
C                   RIGHT HAND SIDE OF THE SYSTEM TO BE SOLVED.
C
C           Y,Z   - N*1 DOUBLE PRECISION  WORK VECTORS.
C
C  OUTPUT:  X     - N*1 DOUBLE PRECISION RESULT VECTOR
C
C
C**********************************************************************
*
      SUBROUTINE SOLVE(NDIM,N,A,P,X,B,Y,Z)
      INTEGER N,NDIM
      DOUBLE PRECISION A(NDIM,N)
      DOUBLE PRECISION X(N),B(N)
      INTEGER P(N)
      DOUBLE PRECISION Y(N),Z(N)
*
      INTEGER I,J
      DOUBLE PRECISION SUM
C
C     SOLVE LY=PB
C
      Y(1) = B(P(1))/A(1,1)
      DO 20 I=2,N
         SUM=0.0
         DO 10 J=1,I-1
            SUM=SUM + (A(I,J)*Y(J))
 10      CONTINUE
         Y(I) = (B(P(I)) - SUM) / A(I,I)
 20   CONTINUE
C
C     SOLVE LTRANSPOSE Z = Y
C
      Z(N) = Y(N)/A(N,N)
      DO 40 I=N-1,1,-1
         SUM=0.0
         DO 30 J=I+1,N
            SUM = SUM + (A(J,I) * Z(J))
 30      CONTINUE
         Z(I) = (Y(I) - SUM) / A(I,I)
 40   CONTINUE
C
C     X = PTRANSPOSE Z
C
      DO 50 I=1,N
         X(P(I))=Z(I)
 50   CONTINUE
*
      RETURN
      END
*
*
*
